#### Application: Task Switching Data - analyse data set

rm(list=ls())

##### Auxiliary functions ##############

### Horicontal direct product of two matrices x.1 and x.2: direct product on all rows of x.1 and x.2, 
### Result is a new matrix by stacking these row vectors into a matrix.
hdir = function(x.1,x.2){ 
  x.1 = as.matrix(x.1)
  x.2 = as.matrix(x.2)
  vers = 0
  if(nrow(x.1)==nrow(x.2)){vers = 1}
  if(nrow(x.2)==1){vers = 2}  
  if(vers==1){
    er = x.1[,rep(seq_len(ncol(x.1)),each=ncol(x.2))]*x.2[,rep.int(seq_len(ncol(x.2)),ncol(x.1))]
  }
  if(vers==2){
    er = x.1%x%matrix(x.2,nrow=1) 
  }
  if(vers==0){
    er = NA
    print("ERROR in hdir: Matrices have not same number of rows")
  }else{
    colnames(er) = paste(rep(colnames(x.1),each=ncol(x.2)),rep.int(colnames(x.2),ncol(x.1)),sep="*")}
  return(er)
}


### Generates design matrix for a saturated design, asymmetric or symmetric restrictions  
### Input: K - vector, legth(K)= no. of factors, whose elements give the no. of factor levels
###        res: 0 - asymmetric restrictions, 
###             1 - symmetric restrictions
###             otherwise identity matrix
###        Fames: names of elements in K 
m.eff = function(K,ia=0,res=0,Fnames=NA){
  x = c()
  cond = c()
  if(res==0){
    if(length(Fnames)==1){if(is.na(Fnames)){Fnames = LETTERS[1:length(K)]}}
    ### asymmetric restrictions
    for(k in 1:length(K)){
      if(k==1 & length(K)==1){x = cbind(x,diag(K)[,-1]); cond = matrix(seq_len(K),ncol=1); 
      colnames(x) = paste(Fnames[1],1:(K[1]-1),sep=".")}    
      if(k==1 & length(K)>1){
        x = kronecker(diag(K[k])[,-1],matrix(1,prod(K[-k]),1))
        colnames(x) = paste(Fnames[1],1:(K[1]-1),sep=".")
        cond = kronecker(seq_len(K[1]),matrix(1,prod(K[-1]),1))} 
      if(k>1  & k<length(K)){
        temp = kronecker(matrix(1,prod(K[1:(k-1)]),1),diag(K[k])[,-1])
        temp = kronecker(temp,matrix(1,prod(K[(k+1):length(K)]),1))
        colnames(temp) = paste(Fnames[k],1:(K[k]-1),sep=".")
        if(ia==0){x = cbind(x,temp)}else{x = cbind(x,temp,hdir(x,temp))}
        cond = cbind(cond,kronecker(
          kronecker(matrix(1,prod(K[1:(k-1)]),1),seq_len(K[k])),
          matrix(1,prod(K[(k+1):length(K)]),1)))}
      if(k>1  & k==length(K)){
        temp = kronecker(matrix(1,prod(K[1:(length(K)-1)]),1),diag(K[k])[,-1])
        colnames(temp) = paste(Fnames[k],1:(K[k]-1),sep=".")        
        if(ia==0){x = cbind(x,temp)}else{x = cbind(x,temp,hdir(x,temp))}
        cond = cbind(cond,kronecker(matrix(1,prod(K[1:(length(K)-1)]),1),seq_len(K[k])))          
      }
    }
  }
  if(res==1){
    if(length(Fnames)==1){if(is.na(Fnames)){Fnames = LETTERS[1:length(K)]}}    
    ### symmetric restrictions
    for(k in 1:length(K)){
      if(k==1 & length(K)==1){
        temp = as.matrix(diag(K[1])[,-K[1]])
        temp[nrow(temp),] = -1
        x = cbind(x,temp) 
        cond = matrix(seq_len(K[1]),ncol=1)
        colnames(x) = paste(Fnames[1],1:(K[1]-1),sep=".")}    
      if(k==1 & length(K)>1){
        x = as.matrix(diag(K[k])[,-K[k]])
        x[nrow(x),] = -1
        x = kronecker(x,matrix(1,prod(K[-k]),1))
        colnames(x) = paste(Fnames[1],1:(K[1]-1),sep=".")
        cond = kronecker(seq_len(K[1]),matrix(1,prod(K[-1]),1))}
      if(k>1  & k<length(K)){
        temp = as.matrix(diag(K[k])[,-K[k]])
        temp[nrow(temp),] = -1
        temp = kronecker(matrix(1,prod(K[1:(k-1)]),1),temp)
        temp = kronecker(temp,matrix(1,prod(K[(k+1):length(K)]),1))
        colnames(temp) = paste(Fnames[k],1:(K[k]-1),sep=".")
        if(ia==0){x = cbind(x,temp)}else{x = cbind(x,temp,hdir(x,temp))}
        cond = cbind(cond,kronecker(
          kronecker(matrix(1,prod(K[1:(k-1)]),1),seq_len(K[k])),
          matrix(1,prod(K[(k+1):length(K)]),1)))}
      if(k>1 & k==length(K)){
        temp = as.matrix(diag(K[k])[,-K[k]])
        temp[nrow(temp),] = -1
        temp = kronecker(matrix(1,prod(K[1:(length(K)-1)]),1),temp)
        colnames(temp) = paste(Fnames[k],1:(K[k]-1),sep=".")
        if(ia==0){x = cbind(x,temp)}else{x = cbind(x,temp,hdir(x,temp))}
        cond = cbind(cond,kronecker(matrix(1,prod(K[1:(length(K)-1)]),1),seq_len(K[k])))
      }
    }
  }
  if(res!=0 & res!=1){x = diag(prod(K))}else{x = cbind(mu=1,x)}
  return(list(x=x,cond=cond))
} 

### create matrices C for main and interaction effects tests
### Input: K - vector, legth(K)= no. of factors, whose elements give the no. of factor levels
###        Fames: names of elements in K 
gen.C = function(K,ia=1,Fnames=NA){
  if(length(Fnames)==1){if(is.na(Fnames)){Fnames = LETTERS[1:length(K)]}}
  if(ia==1){
    pos.xx = matrix(1,2,1)
    nms = c("mu")    
    for(k in seq_len(length(K))){
      nms = c(nms,Fnames[k])
      tmp = matrix(c(pos.xx[2,ncol(pos.xx)]+1,pos.xx[2,ncol(pos.xx)]+K[k]-1),2,1)
      pos.xx = cbind(pos.xx,tmp)
      if(k > 1){
        for(l in 2:(ncol(pos.xx)-1)){
          tmp.0 = (pos.xx[2,l] - pos.xx[1,l] + 1) * (tmp[2] - tmp[1] + 1)
          pos.xx = cbind(pos.xx,matrix(c(pos.xx[2,ncol(pos.xx)]+1,pos.xx[2,ncol(pos.xx)]+tmp.0),2,1))
          nms = c(nms,paste(nms[l],":",Fnames[k],sep=""))
        }
      }
    }
    tmp = NULL
    tmp[paste("C.",1:ncol(pos.xx),sep="")] = list(NULL)
    tmp[[1]] = matrix(c(1,rep(0,prod(K)-1)),1,prod(K))
    for(l in 2:length(tmp)){
      nr = pos.xx[2,l] - pos.xx[1,l] + 1
      nc = pos.xx[1,l] -1
      if(l < length(tmp)){tmp[[l]] = cbind(matrix(0,nr,nc),diag(nr),matrix(0,nr,prod(K)-nc-nr))}else{
        tmp[[l]] = cbind(matrix(0,nr,nc),diag(nr))}
      names(tmp) = nms
    }
  }else{tmp = NULL}
  return(tmp)
}

### Calulate acceleration parameter for BCa confidence interval 
accel = function(dat){apply(as.matrix(dat),2,function(x){sum((mean(x)-x)**3)/(6*sum((mean(x)-x)**2)**(3/2))})}


#######################################################################################################
############### Start Analyses
#######################################################################################################

rep.anov = 1 # 1: compare with standard repeated measures analysis (compare main and interaction effects with Anova)
             # otherwise: no comparison (asymmetric restrictions)

### read in the data 
data <- read.table("/...directory.../Data.txt",header=TRUE) # this works under Linux; needs possibly to be adapted for other OS
# data contains only the observed valid data 
# id: subject no. 
# k: cell (<- indx2)
# id.k: subject-cell indicator 

attach(data)
data[1:50,]
summary(data)
names(data)
T.ik = tapply(y,list(id,indx2),length) # (n x K) # no. of observations for each unit (=row) in each cell (=column))
summary(T.ik)

## create factor levels (= cond), 
## regular design matrix
###        res: 0 - asymmetric restrictions, 1 - asymmetric restrictions, otherwise identity matrix
###        Fames: names of elements in K 
K = c(length(unique(CTI)),length(unique(Task)),length(unique(Seq)),length(unique(Aff)))
if(rep.anov == 1){Z = m.eff(K,ia=1,res=1,Fnames=c("CTI","Task","Seq","Aff"))
  }else{Z = m.eff(K,ia=1,res=0,Fnames=c("CTI","Task","Seq","Aff"))}  # create design matrix Z
cond = Z$cond
Z = Z$x
colnames(Z)
invZ = solve(Z) # inverse of design matrix Z

n = length(unique(id))
m.y = rowMeans(tapply(y,list(indx2,id),mean))

alpha = 0.05

#### Start simple GEE  (see section 3.2) #################################

est.theta.1 = invZ%*%m.y
est.var.1 = tapply(y,list(indx2,id),mean) - Z%*%matrix(est.theta.1,nrow(est.theta.1),n)
est.var.1 = invZ%*%est.var.1%*%t(est.var.1)%*%t(invZ)/n**2

sch_obj_1 = list(iter=0,est.theta=est.theta.1,est.var=est.var.1)

t.v = est.theta.1/sqrt(diag(est.var.1))
p.v = 2*pnorm(-abs(t.v))
delta = sqrt(diag(est.var.1))*qnorm(1-alpha/2)
erg.1 = cbind(theta=est.theta.1,se=sqrt(diag(est.var.1)),t.value=t.v,ci.lo=(est.theta.1-delta),ci.up=(est.theta.1+delta),pvalue=p.v)
colnames(erg.1) = c("theta","se","t-value","ci.low","ci.up","p-value")
round(erg.1,3)

### End simple GEE

######## Calculate correlations and variances based on simplified GEE estimator (see section 3.1)

temp.y = Z%*%est.theta.1  
temp.y = y - as.vector(temp.y[indx2])

variance = tapply(temp.y,indx2,function(x) sum(x**2)) # K variances
variance = variance/colSums(T.ik) 
temp.y = temp.y*as.vector(sqrt(variance[indx2]))
sr.1 = 0; n.1 = 0
sr.2 = 0; n.2 = 0
for(k in seq_len(prod(K))){
  temp.no = id[indx2==k]
  sr.2 = sr.2 + sum(tapply(temp.y[indx2==k],temp.no,
                           function(x) sum((x%o%x)*(matrix(1,length(x),length(x))-diag(length(x))))/2))
  n.2 = n.2 + variance[k]**2 * sum(T.ik[,k]*(T.ik[,k]-1)/2) 
  if(k < prod(K)){
    temp.0 = temp.y[indx2==k]
    temp.1 = temp.y[indx2>k]
    ttmp.no = id[indx2>k]  
    tempr.1 = 0
    tempn.1 = 0
    for(i in seq_len(n)){
      tempr.1 = tempr.1 + sum(temp.0[temp.no==i]%o%temp.1[ttmp.no==i])
      temp = indx2[id==i & indx2>k]
      temp = tapply(temp,temp,function(x) length(x))
      tempn.1 = tempn.1 + sum((variance[k]*variance[k<(seq_len(length(variance)))])*
                                (length(temp.0[temp.no==i])*temp))
    }
    rm(temp,temp.0,temp.1,ttmp.no)
  }      
  sr.1 = sr.1 + tempr.1
  n.1  = n.1 + tempn.1
  rm(temp.no)
}
sr.2 = sr.2[[1]]/n.2 
sr.1 = sr.1/n.1
rm(n.1,n.2,temp.y)

simple.sr.1 = sr.1
simple.sr.2 = sr.2
simple.variance = variance

rm(variance,sr.1,sr.2)

###############################
### Start iterative simplified GEE ##########

max.it =40; eps = 0.000001;

est.theta.2 = est.theta.1 # start value
iter = 0; krit = 0
erg.iter.v = c()
erg.iter.b = c()
while(iter < max.it & krit==0){
  iter = iter+1
  est.theta.alt = est.theta.2
  temp.y = Z%*%est.theta.2 
  temp.y = y - as.vector(temp.y[indx2])
  
  variance = tapply(temp.y,indx2,function(x) sum(x**2)) # K variances
  variance = variance/colSums(T.ik) 
  temp.y = temp.y*as.vector(sqrt(variance[indx2]))
  sr.1 = 0; n.1 = 0
  sr.2 = 0; n.2 = 0
  for(k in seq_len(prod(K))){
    temp.no = id[indx2==k]
    sr.2 = sr.2 + sum(tapply(temp.y[indx2==k],temp.no,
                             function(x) sum((x%o%x)*(matrix(1,length(x),length(x))-diag(length(x))))/2))
    n.2 = n.2 + variance[k]**2 * sum(T.ik[,k]*(T.ik[,k]-1)/2) 
    if(k < prod(K)){
      temp.0 = temp.y[indx2==k]
      temp.1 = temp.y[indx2>k]
      ttmp.no = id[indx2>k]  
      tempr.1 = 0
      tempn.1 = 0
      for(i in seq_len(n)){
        tempr.1 = tempr.1 + sum(temp.0[temp.no==i]%o%temp.1[ttmp.no==i])
        temp = indx2[id==i & indx2>k]
        temp = tapply(temp,temp,function(x) length(x))
        tempn.1 = tempn.1 + sum((variance[k]*variance[k<(seq_len(length(variance)))])*
                                  (length(temp.0[temp.no==i])*temp))
      }
      rm(temp,temp.0,temp.1,ttmp.no)
    }      
    sr.1 = sr.1 + tempr.1
    n.1  = n.1 + tempn.1
    rm(temp.no)
  }
  sr.2 = sr.2[[1]]/n.2 
  sr.1 = sr.1/n.1
  rm(n.1,n.2,temp.y)
  
  # Calculate \hat{\theta}
  d.ik = T.ik/(1 - sr.2 + (sr.2 - sr.1)*T.ik) # (n x K)

  m.diag = t(d.ik)  # = matrix(t(d.ik),max(indx2),n) # (K x n)
  part.1 = rowSums(m.diag) # = sum_i vec(diag(d_{i,1}, ... , d_{i,K})))
  part.2 = m.diag%*%(t(m.diag)*sr.1/(1+sr.1* colSums(m.diag)))
  tD.invR.D = diag(part.1)-part.2
  
  V = 1/sqrt(variance)
  mean.y = tapply(y,list(indx2,id),mean) # K x n
  mean.y = as.vector(V)*mean.y # (K x n)
  part.1 = rowSums(m.diag*mean.y) # = sum_i vec(diag(d_{i,1}, ... , d_{i,K})))
  part.2 = colSums((t(m.diag)*sr.1/(1+sr.1*colSums(m.diag)))*colSums(m.diag*mean.y))
  tD.invR.y = part.1-part.2
  est.theta.2 = solve(t(Z)%*%diag(V)%*%tD.invR.D%*%diag(V)%*%Z)%*%t(Z)%*%diag(V)%*%tD.invR.y
 
  rm(tD.invR.y,part.1,part.2,mean.y)
   
  if(max(abs(est.theta.2-est.theta.alt)) < eps){krit = 1}
  
  if(krit == 1){# variance estimation
    est.var.2 = tapply(y,list(id,indx2),mean) # (n x K)
    est.var.2 = est.var.2 - matrix(Z%*%est.theta.2,n,ncol(est.var.2),byrow=TRUE)
    est.var.2 = t(est.var.2%*%diag(V)) # (K x n); entspricht V^{-1/2} s^*_i
      
    m.diag = t(d.ik)*est.var.2 # (K x n) 
    part.2 = colSums(m.diag)*sr.1/(1+sr.1*rowSums(d.ik)) # (1 x n)
    part.2 = t(d.ik)%*%diag(part.2) # (K x n)
    part.2 = m.diag - part.2 # (K x n)
    part.2 = part.2%*%t(part.2)
    part.2 = t(Z)%*%diag(V)%*%part.2%*%diag(V)%*%Z    
    
    part.1 = solve(t(Z)%*%diag(V)%*%tD.invR.D%*%diag(V)%*%Z)
    est.var.2 = part.1%*%part.2%*%part.1
    
    rm(m.diag,part.1,part.2)
  }
  
  rm(tD.invR.D)
}
### End iterative simplified GEE ##########

t.v = est.theta.2/sqrt(diag(est.var.2))
p.v = 2*pnorm(-abs(t.v))
delta = sqrt(diag(est.var.2))*qnorm(1-alpha/2)
erg.2 = cbind(theta=est.theta.2,se=sqrt(diag(est.var.2)),t.value=t.v,
            ci.lo=(est.theta.2-delta),ci.up=(est.theta.2+delta),pvalue=p.v)
colnames(erg.2) = c("theta","se","t-value","ci.lo","ci.up","p-value")
round(erg.2,3)

##############################################################################
##### Wald-Test and tests of single hypotheses based on simplified GEE

n.boot1 =  999
n.boot2 =  999

# create matrices C.mat for testing all main and interaction effects
if(rep.anov != 1){
  C.mat = list(
               Intercept = cbind(1,matrix(0,1,31)),
               ## main effects: CTI
               C.CTI = cbind(matrix(0,3,1),diag(3),matrix(0,3,28)),

               ## Task 
               C.Task = cbind(t(rep(0,4)),1,t(cbind(rep(0,27)))),

               ## Seq (Alternation vs Repetition) 
               C.Seq = cbind(t(rep(0,8)),1,t(cbind(rep(0,23)))),

               ## Aff 
               C.Aff = cbind(t(rep(0,16)),1,t(rep(0,15))), 

               ## interactions: CTI * Task 
               C.CTI.Task = cbind(matrix(0,3,5),matrix(c(1,0,0,0,1,0,0,0,1),3,3,byrow=TRUE),matrix(0,3,24)),

               ## CTI * Seq 
               C.CTI.Seq = cbind(matrix(0,3,9),matrix(c(1,0,0,-1,1,0, 0,-1,1),3,3,byrow=TRUE),matrix(0,3,20)),

               ## Task * Seq
               C.Task.Seq = cbind(matrix(0,1,12),1,matrix(0,1,19)),

               ##  CTI * Aff
               C.CTI.Aff = cbind(matrix(0,3,17),matrix(c(1,0,0,-1,1,0,0,-1,1),3,3,byrow=TRUE),matrix(0,3,12)),

               # Task * Aff  
               C.Task.Aff = cbind(matrix(0,1,20),1,matrix(0,1,11)),

               # Seq * Aff
               C.Seq.Aff = cbind(matrix(0,1,24),1,matrix(0,1,7)),

               ## interaction CTI * Task * Seq 
               C.CTI.Task.Seq = cbind(matrix(0,3,13),diag(3),matrix(0,3,16)),

               # CTI * Task  * Aff 
               C.CTI.Task.Aff = cbind(matrix(0,3,21),diag(3),matrix(0,3,8)),

               # CTI * Seq  * Aff 
               C.CTI.Seq.Aff = cbind(matrix(0,3,25),matrix(c(1,0,0,-1,1,0,0,-1,1),3,3,byrow=TRUE),matrix(0,3,4)),

               #  Task * Seq  * Aff 
               C.Task.Seq.Aff = cbind(matrix(0,1,28),1,matrix(0,1,3)),

               # interaction Zeit * Aufg  * Kongr Typ 
               C.CTI.Task.Seq.Aff = cbind(matrix(0,3,29),diag(3)))
}else{
   ## alternatively - comparison with ANOVA
   C.mat = gen.C(K,ia=1,Fnames=c("CTI","Task","Seq","Aff"))
}

zeta = lapply(C.mat,function(x)rep(0,nrow(x)))  # tests are for zeta = 0

# Test: Linear hypotheses 
### chi^2-statistic  #################################################################################
Wald = sapply(seq_len(length(C.mat)),function(x){
              tmp = solve(C.mat[[x]]%*%est.var.1%*%t(C.mat[[x]])) 
              tmp = t(C.mat[[x]]%*%est.theta.1 - zeta[[x]])%*%tmp%*%(C.mat[[x]]%*%est.theta.1 - zeta[[x]])})

names(Wald) = names(C.mat)
crit.value = lapply(seq_len(length(C.mat)),function(x){qchisq(1-alpha,df=nrow(C.mat[[x]]))})
Wald = round(cbind(Wald=unlist(Wald),chisq=unlist(crit.value)),3)

### confidence interval for linear combinations

ci.a = lapply(seq_len(length(C.mat)),function(x){sqrt(diag(C.mat[[x]]%*%est.var.1%*%t(C.mat[[x]])))})
ci.m = lapply(seq_len(length(C.mat)),function(x){C.mat[[x]]%*%est.theta.1})
onesided = lapply(seq_len(length(C.mat)),function(x){
                  ones.lo = C.mat[[x]]%*%est.theta.1-ci.a[[x]]*qnorm(1-alpha)
                  ones.up = C.mat[[x]]%*%est.theta.1-ci.a[[x]]*qnorm(alpha)
                  tmp = cbind(-Inf,round(ones.up,3),round(ci.m[[x]],3),round(ones.lo,3),Inf)})

twosided = lapply(seq_len(length(C.mat)),function(x){
                  LoUp = cbind(C.mat[[x]]%*%est.theta.1-ci.a[[x]]*qnorm(1-alpha/2),round(ci.m[[x]],3),
                               C.mat[[x]]%*%est.theta.1+ci.a[[x]]*qnorm(1-alpha/2))})

for(s in seq_len(length(C.mat))){
    colnames(onesided[[s]]) = c("lo","up","est","lo","up")
    colnames(twosided[[s]]) = c("lo","est","up")}


names(onesided) = names(C.mat)
names(twosided) = names(C.mat)

###### Bootstrap
set.seed(4695783)

mean.y = tapply(y,list(indx2,id),mean) # K x n

boot.Wald = c()
boot.std = c()
boot.lin.b = c()
temp.ci.b = c()

b = 1
for(b in seq_len(n.boot1)){
  boot.m.y  = t(apply(mean.y,1,function(x) x[sample(n,size=n,replace=TRUE)])) # naive bootstrap
  boot.theta = invZ%*%rowMeans(boot.m.y)

  boot.tmp.v = boot.m.y - matrix(Z%*%boot.theta,ncol(Z),n)
  boot.tmp.v = boot.tmp.v%*%t(boot.tmp.v)/(n-1)    
  boot.tmp.v = invZ%*%boot.tmp.v%*%t(invZ)/n

  # collect individual recentered/standardized bootstrap estimates, assuming independence
  boot.std = rbind(boot.std,t((boot.theta-est.theta.1))/sqrt(diag(boot.tmp.v)) ) # repetitions standardized

  boot.W = lapply(seq_len(length(C.mat)),function(x){
    C.mat[[x]]%*%boot.tmp.v%*%t(C.mat[[x]])}) # calculate "denominator" of bootstrap Wald statistic T_W
  
  # for linear hypotheses, independence assumption about linear combinations of boostrap theta estimates
  # collect linear combinations standardized of recentered bootstrap estimates, assuming independence
  boot.lin.b = lapply(seq_len(length(C.mat)),function(x){ 
    rbind(boot.lin.b[[x]],t(C.mat[[x]]%*%(boot.theta - est.theta.1))/sqrt(diag(boot.W[[x]])))}) # see last paragraph section 4
  
  # for Wald statistics: inverse of C^T \hat{Var}(\hat{theta}) C, generated assuming independence
  # collect bootstrap-Wald statistics
  boot.Wald = lapply(seq_len(length(C.mat)),function(x){
    tmp.W = solve(boot.W[[x]])
    c(boot.Wald[[x]],t(C.mat[[x]]%*%(boot.theta - est.theta.1))%*%tmp.W%*%(C.mat[[x]]%*%(boot.theta - est.theta.1)))})
  rm(boot.m.y,boot.theta,boot.tmp.v,boot.W)
}

# boot.Wald: bootstrap Wald statistics
Wald.b = lapply(seq_len(length(C.mat)),function(x) round(quantile(boot.Wald[[x]],probs=1-alpha,type=8),3)) 
Wald = cbind(Wald,BC_a=unlist(Wald.b))
rm(boot.Wald)

### Start --- Bias corrected accelerated Bootstrap: BC_a
# centered parameter estimator
if(ncol(boot.std)>1){
  z.std.0 = qnorm(rowMeans(apply(boot.std,1,function(x) as.numeric(x < 0)))) # standardized values, normal distribution
}else{
  z.std.0 = qnorm(rowMeans(as.numeric(boot.std < 0)))    
}
# standardized values, linear hypothesis
z.lin.0 = lapply(seq_len(length(C.mat)),function(x){
                 tmp =  boot.lin.b[[x]]
                 if(ncol(tmp)>1){qnorm(rowMeans(apply(boot.lin.b[[x]],1,function(x) as.numeric(x < 0))))}else{
                   qnorm(mean(as.numeric(tmp < 0)))}})

## collect bootstrap estimates to calculate the acceleration parameter a (see Efron and Tibshirani, 1993)
temp.std.bb = c() # collect recentered and individually standardized theta estimates   
temp.lin.b = NULL # collect results for standardized linear combinations of recentered estimates
temp.lin.b[paste("temp.lin.b",names(C.mat))] = list(NULL)

for(j in seq_len(n.boot2)){
   boot.m.y  = t(apply(mean.y,1,function(x) x[sample(n,size=n,replace=TRUE)])) # Anzahl Parameter x n

   temp.b = invZ%*%rowMeans(boot.m.y)
   temp.v = boot.m.y - matrix(Z%*%temp.b,ncol(Z),ncol(boot.m.y))

   temp.v = temp.v%*%t(temp.v)/(n-1)    
   temp.v = invZ%*%temp.v%*%t(invZ)/n

   temp.std.bb = rbind(temp.std.bb,as.vector((temp.b-est.theta.1)/sqrt(diag(temp.v)))) # standardized values

   # for linear hypotheses, hypotheses tests
   temp.lin.b = lapply(seq_len(length(C.mat)),function(x){
                      temp.v = C.mat[[x]]%*%temp.v%*%t(C.mat[[x]])
                      temp.b = t(C.mat[[x]]%*%(temp.b-est.theta.1))/sqrt(diag(temp.v)) # standardized version
                      rbind(temp.lin.b[[x]],temp.b)})
   
}
rm(boot.m.y,temp.b,temp.v)

# acceleration parameter for BC_a bootstrap
a = lapply(seq_len(length(C.mat)),function(x) accel(temp.lin.b[[x]])) # linear hypothesis - standardized version
a.std.bb = accel(temp.std.bb) # standardized single parameters

# Confidence intervals
# for linear combinations

## two-sided confidence intervals, see Section 4
temp = lapply(seq_len(length(C.mat)),function(x){
  Lo = pnorm(z.lin.0[[x]] + (z.lin.0[[x]] + qnorm(alpha/2))/(1- a[[x]]*(z.lin.0[[x]] + qnorm(alpha/2)))) # standardized version BCA
  Up = pnorm(z.lin.0[[x]] + (z.lin.0[[x]] + qnorm(1-alpha/2))/(1- a[[x]]*(z.lin.0[[x]] + qnorm(1-alpha/2)))) # standardized version BCA
  temp = vapply(1:ncol(boot.lin.b[[x]]),function(y){quantile(boot.lin.b[[x]][,y],probs=c(Lo[y],Up[y]),type=8)},
                FUN.VALUE=numeric(2))
  Lo = C.mat[[x]]%*%est.theta.1 - sqrt(diag(C.mat[[x]]%*%est.var.1%*%t(C.mat[[x]])))*temp[2,]
  Up = C.mat[[x]]%*%est.theta.1 - sqrt(diag(C.mat[[x]]%*%est.var.1%*%t(C.mat[[x]])))*temp[1,]
  cbind(Lo,Up)
})
# standardized version BCA
bca.lin.n.b = lapply(seq_len(length(C.mat)),function(x)(cbind(lo=round(temp[[x]][,1],3),up=round(temp[[x]][,2],3))))
names(bca.lin.n.b) = names(C.mat)
rm(temp)

# single parameters: standardized
Lo = pnorm(z.std.0 + (z.std.0 + qnorm(alpha/2))/(1- a.std.bb*(z.std.0 + qnorm(alpha/2))))
Up = pnorm(z.std.0 + (z.std.0 + qnorm(1-alpha/2))/(1- a.std.bb*(z.std.0 + qnorm(1-alpha/2))))
temp = sapply(1:ncol(boot.std),function(x){quantile(boot.std[,x],probs=c(Lo[x],Up[x]),type=8)})
Lo = est.theta.1 - sqrt(diag(est.var.1))*temp[2,]
Up = est.theta.1 - sqrt(diag(est.var.1))*temp[1,]
bca.std.n.b = cbind(lo=Lo,up=Up)

## one-sided confidence intervals  
# linear hypotheses
temp = lapply(seq_len(length(C.mat)),function(x){
  Lo = pnorm(z.lin.0[[x]] + (z.lin.0[[x]] + qnorm(alpha))/(1- a[[x]]*(z.lin.0[[x]] + qnorm(alpha)))) # standardized version BCA
  Up = pnorm(z.lin.0[[x]] + (z.lin.0[[x]] + qnorm(1-alpha))/(1- a[[x]]*(z.lin.0[[x]] + qnorm(1-alpha)))) # standardized version BCA
  temp = vapply(1:ncol(boot.lin.b[[x]]),function(y){quantile(boot.lin.b[[x]][,y],probs=c(Lo[y],Up[y]),type=8)},
                FUN.VALUE=numeric(2))
  Lo = C.mat[[x]]%*%est.theta.1 - sqrt(diag(C.mat[[x]]%*%est.var.1%*%t(C.mat[[x]])))*temp[1,]
  Up = C.mat[[x]]%*%est.theta.1 - sqrt(diag(C.mat[[x]]%*%est.var.1%*%t(C.mat[[x]])))*temp[2,]
  cbind(Lo,Up)
})
bca.lin.ones.n.b = lapply(seq_len(length(C.mat)),function(x)(cbind(lo=-Inf,up=round(temp[[x]][,1],3),
                                                                   lo=round(temp[[x]][,2],3),up=Inf)))
names(bca.lin.ones.n.b) = names(C.mat)

# single parameters: standardized  
Lo = pnorm(z.std.0 + (z.std.0 + qnorm(alpha))/(1- a.std.bb*(z.std.0 + qnorm(alpha))))
Up = pnorm(z.std.0 + (z.std.0 + qnorm(1-alpha))/(1- a.std.bb*(z.std.0 + qnorm(1-alpha))))
temp = sapply(1:ncol(boot.std),function(x){quantile(boot.std[,x],probs=c(Lo[x],Up[x]),type=8)})
Lo = est.theta.1 - sqrt(diag(est.var.1))*temp[1,] 
Up = est.theta.1 - sqrt(diag(est.var.1))*temp[2,] 
bca.std.n.ones.b = cbind(lo=-Inf,up=Lo,lo=Up,up=Inf)

rm(Lo,Up,temp,temp.lin.b,a,a.std.bb,z.std.0,z.lin.0)
  
colnames(bca.std.n.b) = c("BCa.lo","BCA.up")
erg.1 = cbind(erg.1,bca.std.n.b)
  
### Ende --- Bias corrected acceletared Bootstrap: BC_a


####### Output ######################################################################################

print("---- Simple GEE estimation:")
print(round(erg.1,4))  # cf. supplementary material, Table 4, set rep.anov == 0

#print("Simple GEE estimation, 1-sided BCa confidence intervals:")
#print(round(bca.std.n.ones.b,3))

print(round(c(simple.sr.1 = simple.sr.1, simple.sr.2 = simple.sr.2),3))
print("simple.variance")
print(sort(round(simple.variance,2)))


#print("----- Simplified GEE estimation")
#print(round(erg.2,3)) 

#print(round(c(sr.1 = sr.1, sr.2 = sr.2),3))
#print("variance")
#print(variance)

################################################################################################################

### Wald test:  (cf. Table 5, main text, set rep.anov == 0)
Wald

## Two-sided confidence intervals assuming normality 
#twosided

## One.sided confidence interval assuming normality (cf. Table 6, main text, set rep.anov == 0)
onesided

## Linear combination, BCa, 2-sided confidence interval:
#bca.lin.n.b

## Linear combination, BCa, 1-sided confidence intervals:  (cf. Table 6, mein text, set rep.anov == 0)
bca.lin.ones.n.b

# compare with Repeated measures ANOVA
library(car) # for traditional within ANOVA

if(rep.anov == 1){
  # generate data set tmp.dat to be used in Anova 
  tmp.dat.y = matrix(tapply(y,list(indx2,id),mean),n*prod(K))  # calculate cell means for each i
  cat.pred =  apply(matrix(1,n,1)%x%cond,2,factor) # declare factor-variables
  
  tmp.dat = as.data.frame(cbind(rep(1:n,each=prod(K)),cat.pred)) # create data frame with id and factor variables; 
                                                                 # dim: n*prod(K) x no. of factors
  
  nms = c("CTI","Task","Seq","Aff")
  names(tmp.dat) = c("id",nms)
  tmp.dat$y = tmp.dat.y 
  rm(tmp.dat.y,cat.pred)
  
  ## Univariate main and interaction effects, sphericity tests, GG and HF corrections
  t.dat.new = t(matrix(tmp.dat$y,prod(K),n)) # dim: n x K
  tmp.nms = sapply(1:ncol(cond),function(x)paste(nms[x],cond[,x],sep=""))
  tmp.nms = sapply(1:nrow(tmp.nms),function(x) paste(tmp.nms[x,],collapse="."))
  colnames(t.dat.new) = tmp.nms
  m.1 = lm(t.dat.new ~ 1)
  
  nms.idata = nms
  idata = vector("list", length(nms.idata))   
  names(idata) = nms.idata
  rm(nms.idata)
  for(j in 1:length(K)){
    if(j==1){idata[[1]] = factor(rep(paste(nms[1],"",1:K[1],sep=""),each=prod(K[-1])))}
    if(1<j & j<length(K)){tmp = rep(paste(nms[j],"",1:K[j],sep=""),each=prod(K[-(1:j)]))
    idata[[j]] = factor(rep(tmp,times=prod(K[1:(j-1)])))}
    if(j==length(K)){idata[[j]] = factor(rep(paste(nms[j],"",1:K[j],sep=""),times=prod(K[-j])))}
  }
  idata = data.frame(idata)
  fmla = as.formula(paste("~",paste(colnames(idata),collapse="*")))
  m.2 = Anova(m.1,idata=idata,idesign= fmla, type=3)
  m.2 = summary(m.2, multivariate=FALSE)
  
  rownames(Wald)[rownames(Wald) == "mu"] = "(Intercept)"
  Wald = data.frame(Wald)
  anova.mod = data.frame(m.2$univariate.tests[1:16,1:6])
  aov.adjust = data.frame(m.2$pval.adjustments[1:8,1:4])
  sphericity = data.frame(m.2$sphericity.tests[1:8,1:2])
  
  result = c()
  result = merge(round(anova.mod[,5:6],3),round(sphericity,3),by = "row.names",all.x=TRUE)
  rownames(result) = result[,1]
  result = result[,-1]
  result = merge(result,round(aov.adjust,3),by = "row.names",all.x=TRUE)
  rownames(result) = result[,1]
  result = result[,-1]
  result = merge(result,Wald,by = "row.names")
  colnames(result)[c(3,4,5,7,9)] = c("P.F","sph.stat","P.sph","P.F.GG","P.F.HF")
  result # (cf. supplementary material, Table 3) 
}
